(*RUN THIS FIRST
IMPLEMENTATION ORDER:
Utilities.ml
LoadFeatures.ml
ProbabilityModel.ml
Nivre.ml
Printers.ml
pAndR.ml
Surprisal.ml
MAIN.ml
*)

(**Probability loading. *)

(** {% \subsection{Preliminaries} %} *)
open Utilities;;

(*REMOVE*)
(*#load "str.cma";;*)
(*/REMOVE*)

(*I use this to debug the parser on a PC (yacc and lex won't work)*)
(*REMOVE ALL*)
(*
(*load test sentences into 2 lists: SentenceList and WordList*)
let inFile2 = open_in "../Input/PSCCompare.txt";;

let rec inputWordList inFile2 wordList sentenceList i = try (match input_line inFile2 with
  "" -> (inputWordList inFile2 [] (List.rev wordList::sentenceList) 0)
  |s -> 
     let wordParts = (Str.split (Str.regexp "[ \t]+") s) in
     let node = try({word=(List.nth wordParts 0);lemma="";cat=(List.nth wordParts 1);pos=i;parent=(int_of_string (List.nth wordParts 2));toPeek=true})
     with Invalid_argument "List.nth" -> {word="";lemma="";cat="";pos=i;parent=(-2);toPeek=true} in
       (inputWordList inFile2 (node::wordList) sentenceList (i+1)))
  with End_of_file -> if (List.length wordList > 0) then (List.rev ((List.rev wordList)::sentenceList)) else (List.rev sentenceList);;
*)
(*REMOVE ALL*)

(*REMOVE*)
(*let sentenceList = inputWordList inFile2 [] [] 0;;*)
(*/REMOVE*)


(**These are the data structures that are used to gather probability information for each feature.*)
type genNode = {genName: string; genProb: float};;

(** {i varProb} holds the probabilities for each action.  It is the element in a hashtable where the key is the feature string.*)
type varProb = {left: float; right: float; shift: float; reduce: float; genPOS: genNode list};;

(**Each feature is stored in its own hashtable. The key is the string for the feature (e.g. NN**VBD for Stack1) and the element is a varProb.*)
let top = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let next = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let distance = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let position = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let topChildren = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let topParent = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let topLeft = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let topRight = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack1 = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stackSize = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack3 = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack2 = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack1Next = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack2Next = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;
let stack3Next = (Hashtbl.create 100000 : (string, varProb) Hashtbl.t);;

(**This just converts strings to the hash tables.*)
let findHash key = match key with
	 "STACK3" -> stack3
       |"STACK1" -> stack1
       |"DISTANCE" -> distance
       |"NEXT" -> next
       |"POSITION" -> position
       |"TOP" -> top
       |"TOPCHILDREN" -> topChildren
       |"TOPLEFT" -> topLeft
       |"TOPRIGHT" -> topRight
       |"TOPPARENT" -> topParent
       |"STACKSIZE" -> stackSize
       |"STACK2" -> stack2
       |"STACK3NEXT" -> stack3Next
       |"STACK2NEXT" -> stack2Next
       |"STACK1NEXT" -> stack1Next
       |_ -> topChildren;;

(*REMOVE*) 
(*let weightsIn = (open_in "../Input/Weights/negraTigerPOSStacksNext.txt");;*)
(*/REMOVE*)


(** {% \subsection{Variable loading} %} *)  

(**This function inputs the probabilities into their respective hashtables.*)
let rec inputWeights inFile = try (match (input_line inFile) with
    "" -> ""
  |s -> let line =  (Str.split (Str.regexp "[ \t]+") s) in
     let feature= (List.nth line 0) in
     let weightLeft = (try (float_of_string (List.nth line 1)) with Failure "float_of_string" -> 0.0 | Failure "nth" -> 0.0) in
     let weightRight = (try (float_of_string (List.nth line 2)) with Failure "float_of_string" -> 0.0 | Failure "nth" -> 0.0) in
     let weightReduce = (try (float_of_string (List.nth line 3)) with Failure "float_of_string" -> 0.0 | Failure "nth" -> 0.0) in 
     let weightShift = (try (float_of_string (List.nth line 4)) with Failure "float_of_string" -> 0.0 | Failure "nth" -> 0.0) in
     let keyW = (Str.split (Str.regexp "_") feature) in
     let key1 = (List.hd keyW) in
     let key22 =try (List.nth keyW 1) with Failure "nth" -> "" in
     let key2 = (if (key22 = "") then "NULL" else key22) in
     let hash = (findHash key1) in
     let newVP = ({left=weightLeft;right=weightRight;reduce=weightReduce;shift=weightShift;genPOS=[]}) in
     let test = (Hashtbl.add hash key2 newVP) in 
	    (inputWeights inFile))
  with End_of_file -> "";;

(*REMOVE*)
(*let inWeights = inputWeights weightsIn;;*)
(*/REMOVE*)

(*Load in the lexicalized weights*)
(*REMOVE*)
(*let posIN = (open_in "../Input/Weights/negraTigerGenPOSMap.txt");;
let wordIN = (open_in "../Input/Weights/negraTigerGenWordMap.txt");;*)
(*REMOVE*)

(** {% \subsection{Generative probabilities} %} *)
(** The parser currently doesn't operate on generative probabilities, mainly because the feature files are not included with\
 this distribution.  But, the parser is coded to work for this, so it's just a matter of creating them and then choosing the -generative flag at the command line.*)


(**For generative probabilities, the different Shifts (Shift_NN, Shift_VV) are mapped to integers.  This next function translates it all back.*)
let rec inputArray inFile array = 
  try (match (input_line inFile) with
	   "" -> ""
	 |s -> let line = (Str.split (Str.regexp "[ \t]+") s) in
            let intVal = try (int_of_string (List.nth line 0)) with Failure "nth" -> -1 in
            let key = try (List.nth line 1) with Failure "nth" -> "" in
	    let add = (array.(intVal) <- key) in
            (inputArray inFile array))
  with End_of_file -> "";;

(**For generative probabilities*)
let rec inputWordHash inFile hash =
  try (match (input_line inFile) with
	   "" -> ""
	 |s -> let line = (Str.split (Str.regexp "[ \t]+") s) in
	    let intVal = try (int_of_string (List.nth line 0)) with Failure "nth" -> -1 in
            let key = try (List.nth line 1) with Failure "nth" -> "" in
	    let add = (Hashtbl.add hash key intVal) in
	      (inputWordHash inFile hash))
  with End_of_file -> "";;

let posArray = (Array.make 100 "");;
let wordHash = (Hashtbl.create 100000: (string, int) Hashtbl.t);;

(*REMOVE*)
(*let inPosArray = inputArray posIN posArray;;
let inLexArray = inputWordHash wordIN wordHash;;*)
(*REMOVE*)

(*This finds the POS probabilities conditioned on the features *)
(*REMOVE*)
(*let featurePOSIn = (open_in "../Input/Weights/negraTigerGenPOSPOSStacksDIY.txt");;*)
(*REMOVE*)

(**This function runs the generative probability loader.  The parser currently doesn't operate on generative probabilities, mainly because the feature files are not included with this distribution.  But, the parser is coded to work for this, so it's just a matter of creating them and then choosing the -generative flag at the command line.*)
let rec addGenProbs line newList = match line with
    [] -> []
  |h::t ->  
     let split = (Str.split (Str.regexp "[|]+") h) in
     let name = try (List.hd split) with Failure "hd" -> "" in
     let prob = try (float_of_string (List.nth split 1)) with Failure "nth" -> 0.0 | Failure "float_of_string" -> 0.0 in
         {genName=name;genProb=prob}::(addGenProbs t newList );;

(**If we're running generative probabilities only on POS (rather than lexical items), this is is how we load them.*)
let rec featureToPOS inFile = try (match (input_line inFile) with
    "" -> ""
  |s -> let line =  (Str.split (Str.regexp "[ \t]+") s) in
     let feature= (List.nth line 0) in
     let probList= (addGenProbs (List.tl line) []) in
     let keyW = (Str.split (Str.regexp "_") feature) in
     let key1 = (List.hd keyW) in
     let key22 =try (List.nth keyW 1) with Failure "nth" -> "" in
     let key2 = (if (key22 = "") then "NULL" else key22) in
     let hash = (findHash key1) in
     let element = try (Hashtbl.find hash key2) with Not_found -> {left=0.;right=0.;shift=0.;reduce=0.;genPOS=[]} in
     let element2 = {left=element.left;right=element.right;shift=element.shift;reduce=element.reduce;genPOS=probList} in
     let test = (Hashtbl.add hash key2 element2) in 
	    (featureToPOS inFile))
  with End_of_file -> "";;



(*This finds the word probability based on the POS*)
(*REMOVE*)
(*let wordPOSIn = (open_in "../Input/Weights/negraTigerPOSStacksNext.txt");;*)
(*REMOVE*)

let rec findWordWithProb list word nullProb = match list with
    [] -> nullProb
  |h::t -> let pairList =  (Str.split (Str.regexp "[|]+") h) in
    let pairWord = try (List.hd pairList) with Failure "hd" -> "" in
    let prob = try (float_of_string (List.nth pairList 1)) with Failure "nth" -> 0.0 |Failure "float_of_string" -> 0.0 in
      match pairWord with
	  "NULL" -> (findWordWithProb t word prob)
	|word -> prob
	|_ -> (findWordWithProb t word nullProb);;
    

let rec addLexProbs pos novels hash line = match novels with
    [] -> ""
  |h::t -> if (h.cat <> pos) then (addLexProbs pos t hash line) else (
      let prob = (findWordWithProb (List.tl line) h.word 0.0) in
      let add = (Hashtbl.add hash ((String.uppercase h.word)^"|"^(String.uppercase h.cat)) prob) in
	   (addLexProbs pos t hash line));;

let rec posToLexical inFile novels hash = try (match (input_line inFile) with
    "" -> ""
  |s -> let line =  (Str.split (Str.regexp "[ \t]+") s) in
     let pos= (List.nth line 0) in
     let add = (addLexProbs pos novels hash (List.tl line)) in
       (posToLexical inFile novels hash))
  with End_of_file -> "";; 
     
let rec isNovel word novels = match novels with
    [] -> true
  |h::t -> if (((String.lowercase word.word) = (String.lowercase h.word)) & ((String.uppercase word.cat) = (String.uppercase h.cat))) then false else (isNovel word t);;

let rec findIt list string i = match list with
    [] -> -1
  |h::t -> if (h = string) then i else (findIt t string (i+1));;

let rec addNovels wl nl = match wl with
    [] -> nl
  |h::t -> if ((isNovel h nl) = true) then (addNovels t (h::nl)) else (addNovels t nl);;

let rec findNovelWords sl nl = match sl with
    [] -> nl
  |h::t -> (findNovelWords t (addNovels h nl));;

let lexicalizedHash = (Hashtbl.create 10000: (string, float) Hashtbl.t);;

(*REMOVE*)
(*let posInput = featureToPOS featurePOSIn;;
let novelWords = findNovelWords sentenceList [];;
let wordInput = posToLexical wordPOSIn novelWords lexicalizedHash;;*)
(*REMOVE*)


