open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Retrieval;;

(**The actual Nivre implementation.*)

(**This module defines the Nivre transition system.  It creates a new configuration from an old one using 4 transitions: Shift, LeftArc, RightArc, and Reduce.*)

(** Shift:  Make the first word on tau the first word on the s. *)
let shift configuration = match configuration with 
    {stack=s; tau=hTau::tTau;h= h;d= d; soFar=sf; time=t; decay=dec} -> let newTime, newDecay = (findTimeCost configuration Shift (hTau.pos)) in
      {stack=hTau::s;tau=tTau; h=h; d=d; soFar=sf@[hTau]; time=newTime; decay=newDecay}
  |_ -> configuration;;

(** RightArc: Make the first word on tau the dependent of the first word on s. Shift the first word on tau. *)
let rightArc configuration i =  match configuration with
	{stack=s; tau=th::tt; h=h; d=d; soFar=sf; time=t} -> let newTime, newDecay = (findTimeCost configuration Right i) in
	  (match i with
	    -1 -> {stack=(th::s); tau=tt; h=(H.add (Int32.of_int th.pos) (-1) h); d=(D.add (Int32.of_int th.pos) "right" d);soFar=sf@[th]; time=newTime; decay=newDecay}
            |_ ->  {stack=(th::s); tau=tt; h=(H.add (Int32.of_int th.pos) i h); d=(D.add (Int32.of_int th.pos) "right" d); soFar=sf@[th];time=newTime;decay=newDecay})
	|_ -> configuration;;
    
(** LeftArc: Make the first word on s the dependent of the first word on tau. Reduce the first word on s.*)
let leftArc configuration i = match configuration with
    {stack=sh::st; tau=th::tt; h=h; d=d; soFar=sf} ->  let newTime, newDecay = (findTimeCost configuration Left i) in
      (match th.pos with
	-1 -> {stack=st; tau=th::tt; h=(H.add (Int32.of_int i) (-1) h); d=(D.add (Int32.of_int i) "left" d); soFar=sf;time=newTime;decay=newDecay}
      |_ -> {stack=st; tau=th::tt; h=(H.add (Int32.of_int i) th.pos h); d=(D.add (Int32.of_int i) "left" d); soFar=sf;time=newTime;decay=newDecay})
    |_ -> configuration;;
	      
(** Reduce: Pop the first word on s. *)
let reduce configuration = match configuration with 
    {stack=sh::st; tau=t; h=h; d=d; soFar=sf} -> let newTime, newDecay = (findTimeCost configuration Reduce (-1)) in 
      {stack=st; tau=t; h=h; d=d; soFar=sf;time=newTime;decay=newDecay}
  |_ -> (configuration);;


(*let outputH h = H.fold (fun s n l -> (s, n)::l) h [];;*)


(**I don't think these functions are used anymore...investigate later.*)

let rec findHead hl pos = match hl with
    [] -> false
  |(k,e)::t -> let kk = (Int32.to_int k) in
      if (kk == pos) then true else (findHead t pos);;

let checkHead config = match config with
    {stack=sh::st;tau=t;h=h;d=d;soFar=sf} -> 
     (let hlist = (outputH h) in
      let hasHead = (findHead hlist sh.pos) in
      let tLen = (List.length t) in
	if ((hasHead == false) && (tLen > 0)) then (leftArc config sh.pos) else reduce config)
  |_ -> (config);;

let makeArcs config arcs =  
  let a =  (ArcSet.max_elt arcs) in
	  match a.op with
	     Left -> (leftArc config a.toWord.pos)
	   |Right -> (rightArc config a.fromWord.pos)
	   |Reduce -> checkHead config
	   |Shift -> shift config;;
