open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Nivre;;
open Printers;;

(** Parser accuracy output*)

(** This outputs the precision and recall information associated with the parse across all sentences in the input file.  Note that it only gets output if the input file includes the gold-standard parses (which are NOT used during the parse itself). *)

(** Make the output printer. *)
let prPrinter = (open_out "../p&r.txt");;

(**This function gets the word's parent.*)
let rec getParent c h i = match h with
    [] -> (-1)
  |(a,b)::t -> if ((Int32.compare (Int32.of_int i) a) == 0) then 
	(try ((List.nth c.soFar b).pos) 
	  with Invalid_argument "List.nth" -> (-1))
	else (getParent c t i);;

(**This function adds up the total_matched (tm), total_proposed (tp) and total (tot) for a given sentence. Error is a boolean which tells whether there was an error in the sentence, in which case the sentence number will be output. *)
let rec addItAllUp item soFar error tm tp tot= match soFar with
    [] -> tm, tp, tot, error
  |node::t -> match node.cat with
	"$." -> addItAllUp item t error tm tp tot
      |x -> 
	 let parserParent = (getParent item.config (outputH item.config.h) node.pos) in
	 let realParent = node.parent in
            if (parserParent = realParent)
            then (addItAllUp item t error (tm+.1.) (tp+.1.) (tot+.1.))
            else (if (parserParent == (-1))
                 then (addItAllUp item t true tm (tp+.1.) (tot+.1.))
	         else (addItAllUp item t true tm tp (tot+.1.)));;

(** This is the printer driver. *)
let rec runPRPrinter finals i errorList tm tp tot = match finals with
    [] -> tm, tp, tot, (List.rev errorList)
  |h::t -> let item = (List.hd h) in
    let thisTM, thisTP, thisTOT, error = (addItAllUp item item.config.soFar false tm tp tot) in
    if error then (runPRPrinter t (i+1) (i::errorList) thisTM thisTP thisTOT) else (runPRPrinter t (i+1) errorList thisTM thisTP thisTOT)
      
(** This outputs the errors. *)    
let rec outputPRErrors el p = match el with
    [] -> flush p
  |h::t -> ((output_string p ((string_of_int h)^", ")); (outputPRErrors t p));;

(**This does the precision and recall calculation for all sentences plus outputs the info and errors.  In other words, it's the driver for the whole module. *)
let pAndRPrint finals p model =  match ((List.hd (List.hd (List.hd finals)).config.soFar).parent) with
     (-2) -> (flush p)
   |_ -> 
  let totalMatched, totalProposed, totalOutThere, error = (runPRPrinter finals 0 [] 0.0 0.0 0.0) in
  let precision = (match totalMatched with
      0.0 -> 0.0
    |x -> (x /. totalProposed)) in
  let recall = (match totalMatched with 
		    0.0 -> 0.0
		  |y -> (y /. totalOutThere)) in
  let fscore = ((2.0 *. precision *. recall)/.(precision +. recall)) in
  let outputPR = (output_string p (model^"\nTotal matched: "^(string_of_float totalMatched)^"\nTotal proposed: "^(string_of_float totalProposed)^"\nTotal out there: "^(string_of_float totalOutThere)^"\nAverage precision: "^(string_of_float precision)^"\nAverage recall: "^(string_of_float recall)^"\nF-score: "^(string_of_float fscore)^"\nSentences with errors: ")) in
  let outputErrors = (outputPRErrors error p) in 
  ((output_string p "\n\n\n"); (flush p));;

     
      
