open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Nivre;;

(** Various outputs for the parser. *)

(** Because a lot of these are just finicky outputs I have coded for various papers, I'm going to leave the majority uncommented.  I will just give overviews on what each type of printer does. *)


(** {% \subsection{Preliminaries} %} *)

(** Flushing. *)
let rec flushAll p = match p with
    [] -> ignore (1 +1)
  |h::t -> flush h; flushAll t;;

(** Output closing. *)
let rec close_outAll p = match p with
    [] -> ignore (1 + 1)
  |h::t -> close_out h; close_outAll t;;

(** Create a printer specific to a sentence number and a feature. *)
let getPrinter model sentnumber = [(open_out ("../Output/"^model^"-"^(string_of_int sentnumber)^".txt"))];;
let rec getDepPrinters model sentnumber k printers = match k with
    0 -> (List.rev printers)
  |x -> getDepPrinters model sentnumber (k-1) ((open_out ("../Output/"^model^"-"^(string_of_int sentnumber)^"-"^(string_of_int k)^".txt"))::printers);;

let rec getDGPrinters model sentnumber k printers = match k with
    0 -> (List.rev printers)
  |x -> getDGPrinters model sentnumber (k-1) ((open_out ("../Output/"^model^"-"^(string_of_int sentnumber)^"-"^(string_of_int k)^".ps"))::printers);;

let rec getOnePrinter model = (open_out("../Output/"^model^".txt"));;


(** {% \subsection{Config Printer} %} *)
(**This printer outputs configuration information.  It has information specific to each parser state and the transitions between them for a parse.  It's really good for debugging.*)

let rec outputList x s = match s with
    [] -> (output_string x "\n")
  |h::t -> (output_string x h.word); (output_string x " "); (output_string x h.lemma); (output_string x " ");(output_string x h.cat); (output_string x " "); (output_string x (string_of_int h.pos)); (output_string x "::"); outputList x t;;

let rec outputHLists l x = match l with
    [] -> (output_string x "\n")
  |(k,e)::t -> (output_string x (Int32.to_string k)); (output_string x " WITH "); (output_string x (string_of_int e)); (output_string x "::");(outputHLists t x);;

let rec outputDLists l x = match l with
    [] -> (output_string x "\n")
  |(k,e)::t -> (output_string x (Int32.to_string k)); (output_string x " IS "); (output_string x e); (output_string x "::");(outputDLists t x);;

let outputH h = H.fold (fun s n l -> (s, n)::l) h [];;
let outputD d = D.fold (fun s n l -> (s, n)::l) d [];;

let rec outputHistory x hist = match hist with
    [] -> (output_string x "\n")
  |h::t -> (output_string x ((string_of_op h.arcInfo.op)^"-"^(string_of_float h.arcInfo.prob)^"-"^(h.arcInfo.fromWord.word)^"-"^(h.arcInfo.toWord.word)^"-"^(string_of_int h.rank)^";"));
      (outputHistory x t);; 
    

let outputConfigAndProb x h = (output_string x ((string_of_int h.rank)^"\n"));(outputList x h.config.stack); (outputList x h.config.tau); (outputHLists (outputH h.config.h) x); (outputDLists (outputD h.config.d) x); (output_string x (string_of_op h.arcInfo.op)); (output_string x " "); (output_string x (string_of_float h.arcInfo.prob)); (output_string x " "); (output_string x (string_of_int h.arcInfo.fromWord.pos));(output_string x (string_of_int h.arcInfo.toWord.pos)); (output_string x "\n");(output_string x ((string_of_float h.forwardProb.fprob)^"\n"));(output_string x ((string_of_bool h.forwardProb.forSurprisalCalc)^"\n"));(outputHistory x (List.rev h.history)); (output_string x "\n");(output_string x "\n"); (output_string x "\n");;

let rec configPrint x p = match x with
    [] -> (output_string p "\n\n")
  |h::t -> (outputConfigAndProb p h); (configPrint t p);;


(** {% \subsection{Tensor Product Input} %} *)

(** This is a very crude function that can tell you whether the arc the parser took was the arc it should have taken.  It doesn't work too well, though, because in order for it to be perfect I would need a lot more information than is available here.  And that would require more time.  But, it does work semi-ok.*)
let correctParse x s1 t1 h = match h.arcInfo.op with
    Left -> if (h.arcInfo.toWord.parent = t1.pos) then (output_string x "Yes\n") else (output_string x ("No. "^(string_of_int s1.pos)^" with "^(string_of_int s1.parent)^"\n"))
   |Right -> if (h.arcInfo.toWord.parent = s1.pos) then (output_string x "Yes\n") else (output_string x ("No. "^(string_of_int t1.pos)^" with "^(string_of_int t1.parent)^"\n"))
   |Shift|Reduce -> (output_string x "Not sure\n");;

(** The actual output string formatting. *)
let outputTensorAndProb x h = 
  let s1 = try (List.hd h.config.stack) with Failure "hd" -> {word="NULL";lemma="NULL";cat="NULL";pos=(-1);parent=(-1);toPeek=false} in
  let t1 = try (List.hd h.config.tau) with Failure "hd" -> {word="NULL";lemma="NULL";cat="NULL";pos=(-1);parent=(-1);toPeek=false} in
    (output_string x "T: "); (outputHLists (outputH h.config.h) x); 
    (output_string x "p: "); (output_string x (s1.word^"|"^s1.cat^"|"^(string_of_int s1.pos)^"\n"));
    (output_string x "q: "); (output_string x (t1.word^"|"^t1.cat^"|"^(string_of_int t1.pos)^"\n"));
    (output_string x "I: "); (outputList x h.config.tau);
    (output_string x "Transition Probability: "); (output_string x ((string_of_float h.arcInfo.prob)^"\n"));
    (output_string x "Forward Probability: "); (output_string x ((string_of_float h.forwardProb.fprob)^"\n"));
    (output_string x "Position in beam: "); (output_string x ((string_of_int h.rank)^"\n"));
    (output_string x "Action taken: "); (output_string x ((string_of_op h.arcInfo.op)^" "^(string_of_int h.arcInfo.toWord.pos)^" with "^(string_of_int h.arcInfo.fromWord.pos)^"\n"));
    (output_string x "Correct action: "); (correctParse x s1 t1 h);
    (output_string x "History: "); (outputHistory x (List.rev h.history));;

(** The driver. *)
let rec tensorPrint x p = match x with
    [] -> (output_string p "\n\n")
  |h::t -> (outputTensorAndProb p h); (output_string p "\n\n"); (tensorPrint t p);;


(** {% \subsection {k-Best Printer} %} *)

(**This outputs the information for all ks (rather than final-only*)

let rec outputList x s = match s with
    [] -> (output_string x "\n")
  |h::t -> (output_string x h.word); (output_string x " "); (output_string x " "); (output_string x h.lemma); (output_string x h.cat); (output_string x " "); (output_string x (string_of_int h.pos)); (output_string x "::"); outputList x t;;

let rec outputHLists l x = match l with
    [] -> (output_string x "\n")
  |(k,e)::t -> (output_string x (Int32.to_string k)); (output_string x " WITH "); (output_string x (string_of_int e)); (output_string x "::");(outputHLists t x);;

let rec outputDLists l x = match l with
    [] -> (output_string x "\n")
  |(k,e)::t -> (output_string x (Int32.to_string k)); (output_string x " IS "); (output_string x e); (output_string x "::");(outputDLists t x);;

let outputH h = H.fold (fun s n l -> (s, n)::l) h [];;
let outputD d = D.fold (fun s n l -> (s, n)::l) d [];;

let rec outputHistory x hist = match hist with
    [] -> (output_string x "\n")
  |h::t -> (output_string x ((string_of_op h.arcInfo.op)^"-"^(string_of_float h.arcInfo.prob)^"-"^(h.arcInfo.fromWord.word)^"-"^(h.arcInfo.toWord.word)^"-"^(string_of_int h.rank)^";"));
      (outputHistory x t);; 
    
let outputKBest x h = 
  let word = try ((List.hd h.config.tau).word) with Failure "hd" -> "NULL" in 
  (output_string x ("Word: "^word));(output_string x "\n");
  (output_string x "Stack: "); (outputList x h.config.stack); 
  (output_string x "WordList: ");(outputList x h.config.tau);
  (output_string x ("Operation: "^(string_of_op h.arcInfo.op))); (output_string x " "); (output_string x ("Prob: "^(string_of_float h.arcInfo.prob))); (output_string x " "); (output_string x ("From "^(string_of_int h.arcInfo.fromWord.pos)));(output_string x ("To "^(string_of_int h.arcInfo.toWord.pos))); (output_string x "\n");
  (output_string x ("Forward Prob: "^((string_of_float h.forwardProb.fprob)^"\n")));
 (output_string x "History: "); (outputHistory x (List.rev h.history)); (output_string x "\n\n\n");;

let rec kbestPrint x p = match x with
    [] -> (output_string p "\n\n")
  |h::t -> (outputKBest p h); (kbestPrint t p);;

(** {% \subsection{DepPrint Input} %} *)

(**This outputs to a DepPrint input file, which can be used with the graph visualization software DepPrint (available on Marisa's website. *)

let rec outputWords p sf = match sf with
    [] -> flush p
  |h::t -> (output_string p ("<word>\t"^h.word^"\t"^h.cat^"\n")); (outputWords p t);;

let rec outputArcs p hlist = match hlist with
    [] -> flush p
  |(k,e)::t -> if (e == (-1)) then (outputArcs p t) else ((output_string p ("<arc>\t"^(Int32.to_string k)^"\t"^(string_of_int e)^"\n")); (outputArcs p t));;
    
let outputSurps p surpArray until= 
  let surpList = (Array.to_list surpArray) in
  let rec outSurps p sl i until = if (i >= until) then (flush p) else match sl with
      [] -> flush p
    |h::t -> if (h = 0.0) then (outSurps p t (i+1) until) else ((output_string p ("<surprisal>\t"^(string_of_int i)^"\t"^(string_of_float h)^"\n")); (outSurps p t (i+1) until))
  in (outSurps p surpList 0 until);;

let rec outputCompares p soFar = match soFar with
    [] -> flush p
  |h::t -> match h.parent with
	(-1) -> (outputCompares p t)
      |(-2) -> (outputCompares p t)
      |x -> (output_string p ("<arc>\t"^(string_of_int h.pos)^"\t"^(string_of_int x)^"\t#FF0000\tDASHED\n")); (outputCompares p t);;

let rec depPrint x p surpPrint= match x, p with
    [], [] -> flushAll p
  |h::t, ph::pt -> (if ((List.length h.config.soFar) == 0) then (depPrint t pt surpPrint) else ((output_string ph "<sentence>\n"); (outputWords ph h.config.soFar); (outputArcs ph (outputH h.config.h)); (outputCompares ph (h.config.soFar)); (outputSurps ph surpPrint (List.length h.config.soFar));(output_string ph "</sentence>\n"); (depPrint t pt surpPrint)))
  |_, _ -> flushAll p;;

let rec initialDepPrint x p surpPrint = match p with
    [] -> flushAll p
  |h::t -> (if ((List.length x.config.soFar) == 0) then (initialDepPrint x t surpPrint) else ((output_string h "<sentence>\n"); (outputWords h x.config.soFar); (outputArcs h (outputH x.config.h)); (outputCompares h (x.config.soFar)); (outputSurps h surpPrint (List.length x.config.soFar)); (output_string h "</sentence>\n"); (initialDepPrint x t surpPrint)));;


(** {% \subsection{DGGraph Output} %} *)

(**This outputs to Matthias Buch-Kromann's input format for DGGraph*)

let rec outputHeader p inFile = try (match (input_line inFile) with
  s -> (output_string p (s^"\n")); (outputHeader p inFile)) with
    End_of_file -> (close_in inFile); flush p;;

let rec outputDGWords p sf = match sf with 
    [] -> (output_string p ("] def\n\n\n"^"/edges [\n"))
  |h::t -> (output_string p ("("^(h.word)^") ("^(h.cat)^") word\n")); (outputDGWords p t);;


let rec outputDGArcs p hList = match hList with
  [] -> (output_string p "] def\n\n\ndraw end")
  |(hkid,hparent)::t -> (output_string p ((Int32.to_string (Int32.add hkid Int32.one))^" "^(string_of_int (hparent + 1))^" ( ) edget\n")); (outputDGArcs p t);;

let rec dgOut x p = match x, p with
    [], [] -> flushAll p
  |h::t, ph::pt -> (if ((List.length h.config.soFar) == 0) then (dgOut t pt) else ((outputHeader ph (open_in "../Input/DGHeader.txt"));(output_string ph ("(ROOT) (ROOT) word\n")); (outputDGWords ph h.config.soFar);(outputDGArcs ph (outputH h.config.h));(output_string ph "\n\n\n\n"); (dgOut t pt)))
  |_, _ -> flushAll p;;


(** {% \subsection{CoNLL Output} %} *)
(**This is basically word (tab) position (tab) parent.*)

let rec findParent pos hs = match hs with
    [] -> (-2)
  |(k,e)::t -> if ((Int32.to_int k) == pos) then (e) else (findParent pos t);;

let outputFDWords p l config = 
   let configH = (outputH config.h) in
   let rec outFDWs p l ch = match l with
    [] -> flush p
     |h::t -> 
	(let parent = (findParent h.pos configH) in 
       ((output_string p (h.word^"\t"^h.cat^"\t"^(string_of_int parent)^"\n")); (flush p); (outFDWs p t ch)))
   in (outFDWs p l configH);;

let rec fullDep x p = match x with
    [] -> flush p
  |h::t -> (if ((List.length h.config.soFar)==0) then (fullDep t p) else ((outputFDWords p h.config.soFar h.config); (output_string p "\n"); (flush p); (fullDep t p)))


(** {% \subsection{Driver for all printers} %} *)


(*This iterates through the sentences, selects the appropriate printer, and outputs*)
let printIter configs p out surpPrint finalOnly= 
  let xs = if (finalOnly == true) then [(List.nth configs ((List.length configs) -1))] else (configs) in 
  let rec pIter x p out surpPrint = match x with
    [] -> (flushAll p)
  |h::t -> match out with
	"config" -> (try ((configPrint h (List.hd p)); (output_string (List.hd p) "\n|||||||||\n"); (pIter t p out surpPrint)) with Failure "hd" -> (print_string "error printIter"))
      |"tensor" -> (try ((tensorPrint h (List.hd p)); (output_string (List.hd p) "||||||\n\n"); (pIter t p out surpPrint)) with Failure "hd" -> (print_string "error printIter"))
      |"dgGraph" -> (dgOut h p); (flushAll p); (pIter t p out surpPrint)
      |"dependency" -> (try ((fullDep h (List.hd p)); (flushAll p); (pIter t p out surpPrint)) with Failure "hd" -> (print_string "error printIter"))
      |"kbest" -> (try ((kbestPrint h (List.hd p)); (output_string (List.hd p) "\n***\n"); (pIter t p out surpPrint)) with Failure "hd" -> (print_string "error printIter"))
      |_ -> match (List.length h) with
		 1 -> (initialDepPrint (List.hd h) p surpPrint); (flushAll p); (pIter t p out surpPrint)
	       |_ -> (depPrint h p surpPrint); (flushAll p);(pIter t p out surpPrint)
  in (pIter xs p out surpPrint);;




