open Utilities;;
open LoadFeatures;;
open ProbabilityModel;;
open Nivre;;

(** Parser state space printer *)

(** This printer outputs the entire state space for a parse as an SVG file.  

Its x axis is the words as they are parsed and the y axis is the probability of each parser state.  It kind of looks like a stock market output for a particular stock (with the parse going up and down with different probabilities).  It's useful for explaining how the parser works, and was used in Boston et al. (submitted in 2009).  Currently it only works with probabilities on the y axis, although in the paper I also showed retrieval in the y axis.  That was done by hand, though.  It should be an easy-enough update. 

At any rate, because I'm fairly positive I'm the only person who would use this or update it, I'm leaving it pretty uncommented.
*)


let string_of_op x = match x with
Left -> "LEFT"
|Right -> "RIGHT"
|Shift -> "SHIFT"
|Reduce -> "REDUCE";;  

let getStatePrinter model sentnumber = (open_out ("../Output/States/State-"^model^"-"^(string_of_int sentnumber)^".svg"));;
let getStockPrinter model sentnumber = (open_out ("../Output/States/Stock-"^model^"-"^(string_of_int sentnumber)^".svg"));;

let round x = int_of_float (floor (x +. 0.5));;       

(*STATE-SPACE OUTPUT*)
let makeHeader p width height = (output_string p ("<?xml version=\"1.0\" standalone=\"no\"?>
   <!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"
  \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">
  <svg width=\""^(string_of_int width)^"px\" height=\""^(string_of_int height)^"\"
  xmlns=\"http://www.w3.org/2000/svg\" version=\"1.1\">\n\n
  <defs>
  \t<marker id=\"arrowHead\" viewBox=\"0 0 10 10\" refX=\"1\" refY=\"5\" markerUnits=\"strokeWidth\" orient=\"auto\" markerWidth=\"4\" markerHeight=\"3\">
		<polyline points=\"0,0 10,5 0,10 1,5\" fill=\"#000000\" />
	</marker>
  </defs>\n"));;

let blankRect p x y =  (output_string p 
     ("<rect x=\""^(string_of_int x)^"\" y=\""^(string_of_int y)^"\" rx=\"7\" ry=\"7\" width=\"100\" height=\"70\" style=\"fill:#ffffff;stroke:black;stroke-width:3;opacity:1\"/>\n"));;

let rec printBlanks p states x y = if (states <= 0) then (output_string p "\n\n\n") else ((blankRect p x y); (printBlanks p (states-1) (x+110) y));;

let stateRect p x y name fprob tprob = 
  let t1x = (string_of_int (x+3)) in
  let t1y = (string_of_int (y+11)) in
  let t2x = (string_of_int (x+3)) in
  let t2y = (string_of_int (y+39)) in
  let t3x = (string_of_int (x+3)) in
  let t3y = (string_of_int (y+25)) in
  (output_string p 
     ("\t<rect x=\""^(string_of_int x)^"\" y=\""^(string_of_int y)^"\" rx=\"7\" ry=\"7\" width=\"100\" height=\"70\" style=\"fill:#ffffff;stroke:black;stroke-width:1;opacity:1\"/>
     <text x=\""^t1x^"\" y=\""^t1y^"\" width=\"98\" height=\"68\" font-family=\"Palatino Linotype\" font-size=\"10\" text-anchor=\"start\">"^name^"</text>
     <text x=\""^t2x^"\" y=\""^t2y^"\" width=\"98\" height=\"68\" font-family=\"Palatino Linotype\" font-size=\"10\" text-anchor=\"start\">"^fprob^"</text>
     <text x=\""^t3x^"\" y=\""^t3y^"\" width=\"98\" height=\"68\" font-family=\"Palatino Linotype\" font-size=\"10\" text-anchor=\"start\">"^tprob^"</text>\n\n"));;

let roundish x i = 
  let i = (i+1) in
  let s = (string_of_float x) in
  let decimal = (try (String.index s '.') with Not_found -> -1) in
  if (decimal = (-1)) then x else (
    let truncate = (try (String.sub s 0 (decimal+i)) with Invalid_argument "String.sub" -> "") in
      (try (float_of_string truncate) with Failure "float_of_string" -> x));;
  
 
let outputState p config words k x y = 
  let op = (string_of_op config.arcInfo.op) in
  let cat = config.arcInfo.element in
  let name = (op^" "^cat) in
  let fprob = (string_of_float config.forwardProb.fprob) in
  let tprob = (string_of_float (roundish config.arcInfo.prob 3)) in
  (stateRect p x y name fprob tprob);;

let makeArrow p px py x y = (output_string p ("<line x1=\""^(string_of_int px)^"\" x2=\""^(string_of_int x)^"\" y1=\""^(string_of_int py)^"\" y2=\""^(string_of_int y)^"\" style=\"fill:none;stroke:black;stroke-width:5;\" marker-end=\"url(#arrowHead)\"/>\n\n"));; 

let outputArrow p config x y = if (List.length config.history < 2) then (flush p) else (
  let last = try (List.hd config.history) with Failure "hd" -> config in
  let rank = last.rank in
  let parentX = (40 + (100*rank) + (10*(rank-1)) - 50) in
  let parentY = (y - 70) in
  let thisX = (x + 50) in
  let thisY = (y - 10) in
    (makeArrow p parentX parentY thisX thisY));;

let rec statePrint kConfigs p words k x y = match kConfigs with
    [] -> (*let remainingStates = ((k*4)-k) in
          (printBlanks p remainingStates x y)*)
          (flush p)
  |h::t -> 
     let stateOutput = (outputState p h words k x y) in
     let arrowOutput = (outputArrow p h x y) in
    (statePrint t p words k (x+110) y);;

(*Horse race output*)
let dummyNode = {word="NULL";lemma="NULL";cat="NULL";pos=(-2);parent=(-2);toPeek=false};;

let findWhichWord it = match it.config.tau with
    	[] -> (match it.arcInfo.op with 
	    Shift|Right ->(match it.config.stack with
		[] -> dummyNode
	      |hh::tt -> ((List.hd it.config.stack)))
	  |_ -> dummyNode)
       |h::t -> (match it.arcInfo.op with
	   Left|Reduce -> ((List.hd it.config.tau))
	   |Shift|Right -> (match it.config.stack with
		 [] -> dummyNode
	       |hhh::ttt ->((List.hd it.config.stack))));;


let makeLine p px py x y = (output_string p ("<line x1=\""^(string_of_int px)^"\" x2=\""^(string_of_int x)^"\" y1=\""^(string_of_int py)^"\" y2=\""^(string_of_int y)^"\" style=\"fill:none;fill-rule:evenodd;stroke:#b3b3b3;stroke-width:5;\"/>\n\n"));; 

let makeBlackLine p px py x y = (output_string p ("<line x1=\""^(string_of_int px)^"\" x2=\""^(string_of_int x)^"\" y1=\""^(string_of_int py)^"\" y2=\""^(string_of_int y)^"\" style=\"fill:none;fill-rule:evenodd;stroke:#000000;stroke-width:5;\"/>\n\n"));; 

let printText p x y text =  (output_string p ("<text x=\""^(string_of_int x)^"\" y=\""^(string_of_int y)^"\" width=\"98\" height=\"68\" font-family=\"Palatino Linotype\" font-size=\"28\" text-anchor=\"end\">"^text^"</text>"));;

let makeScale p high x = 
  let limit = (round ((float_of_int (high + 100))/.100.)) in
  let top = (high + 500 - (limit*100)) in
  let ml = (makeBlackLine p x (high+500) x top) in
  let yy = (high+500) in
  let xx = (x-20) in
  let rec makeText i limit x y = match (i>limit) with
      true -> (flush p)
    |false -> let text = ("0."^(string_of_int i)) in
      let print = (printText p x y text) in
	(makeText (i+1) limit x (y-100)) in
    (makeText 0 limit xx yy);;

let rec sumPrefixProbs wl sum = match wl with
    [] -> sum
  |h::t -> if (h.forwardProb.forSurprisalCalc) then (sumPrefixProbs t (sum+.h.forwardProb.fprob)) else (sumPrefixProbs t sum);;

let calcSurprisal wl lastPP = 
  let prefixProb = sumPrefixProbs wl 0.0 in
  (abs_float ((log ( lastPP /. prefixProb))/.(log 2.0)));;

let outputWord wl p x lastPP high = 
  let baseline = high in
  let n = try (findWhichWord (List.hd wl)) with Failure "hd" -> dummyNode in
  let printWord = (printText p (x+108) (baseline+100) n.word) in
  let printCat = (printText p (x+108) (baseline+140) n.cat) in
  let surprisal  = (roundish (calcSurprisal wl lastPP) 3) in
  let printSurp = (printText p (x+108) (baseline+180) (string_of_float surprisal)) in
  (makeLine p (x+110) (baseline+180) (x+110) 500);;

let rec howFarAway config i = 
  let last = try (List.hd config.history) with Failure "hd" -> config in
  if (last.forwardProb.forSurprisalCalc) then (i) else (howFarAway last (i+1));;


let rec findSequence elementList long = match elementList with
    [] -> long
  |h::t -> let far = (howFarAway h 1) in
      if (far > long) then (findSequence t far) else (findSequence t long);;

let rec findLongSequence wordArray i long = if (i >= 0) then
  (let element = wordArray.(i) in
   (findSequence element 0)) else (1);;
     

let rec getLastStates history list = match history with
    [] -> (List.rev list)
  |h::t -> if (h.forwardProb.forSurprisalCalc = true) then (List.rev list) else (getLastStates t (h::list));;


let rec compareItem item list = match list with
    [] -> (-2)
  |(h,hx)::t -> if ((item.forwardProb.fprob = h.forwardProb.fprob) && (item.arcInfo.element = h.arcInfo.element) && (item.arcInfo.op = h.arcInfo.op) && (item.arcInfo.prob = h.arcInfo.prob)) then (hx+100) else (compareItem item t);;

let rec matchItems histL tcL newH newT x = match histL with
    [] -> (List.rev newH), newT
  |h::t -> 
    if ((compareItem h tcL)>0) then (matchItems t tcL newH newT x) else (matchItems t tcL (h::newH) ((h,x)::newT) (x-200-100));;


let horseStateAndArrowOut config p k x wordX printLast takenCareOf high = 
  let y = (high-(round ((config.arcInfo.prob)*.1000.0))) in
  let op = (string_of_op config.arcInfo.op) in
  let cat = config.arcInfo.element in
  let name = (op^" "^cat) in
  let fprob = (string_of_float config.forwardProb.fprob) in
  let tprob = (string_of_float (roundish config.arcInfo.prob 3)) in
  let h = (stateRect p x y name fprob tprob) in
  (*let h = (blankRect p x y) in*)
  if (printLast = false) then (flush p) else (
  let last = try (List.hd config.history) with Failure "hd" -> config in
  let lastY = (high-(round ((last.arcInfo.prob)*.1000.0))) in
  let parentY = (lastY+35) in
  let parentX = 
    if ((last.forwardProb.forSurprisalCalc)=true) 
    then (wordX+100)
    else (let hx = (compareItem last takenCareOf) in
	    if (hx > 0) then (hx) else (x - 200)) in
  let thisX = (x-15) in
  let thisY = (y+35) in
    (makeArrow p parentX parentY thisX thisY));;
  


let horseHistory config p k x wordX takenCareOf high =
  let history = (getLastStates config.history []) in 
  if ((List.length config.history) < 2) then (let out = (horseStateAndArrowOut config p k x wordX false takenCareOf high) in ((config,x)::takenCareOf)) else(
  let printThis = (horseStateAndArrowOut config p k x wordX true takenCareOf high) in
  let newHistory, newTC = (matchItems history ((config,x)::takenCareOf) [] ((config,x)::takenCareOf) (x-200-100)) in
  let rec historyOut hist p kk xx wx tco high = match hist with
      [] -> tco
    |h::t -> ((horseStateAndArrowOut h p kk xx wx true takenCareOf high); (historyOut t p kk (xx-200-100) wx tco high)) in
  (historyOut newHistory p k (x-200-100) wordX newTC high));;

let rec horsePrint wordList p k x wordX takenCareOf high = match wordList with
    [] -> takenCareOf
  |h::t -> 
     if (h.forwardProb.forSurprisalCalc = true) 
     then (
       let newTaken = (horseHistory h p k x wordX takenCareOf high) in 
	 (horsePrint t p k x wordX newTaken high))
     else (horsePrint t p k x wordX takenCareOf high);;


let rec printNonHistory wordL tcL pp kk xx lastXX high =
  let todo, takenCareOf = (matchItems wordL tcL [] [] (xx-200-100)) in
  let rec print tdl p k x lastX = (match tdl with
      [] -> ()
    |h::t -> 
       let farAway = (howFarAway h 1) in
       let newX = if (h.forwardProb.forSurprisalCalc) then (x) else (lastX + (200*farAway) + (100*farAway)) in
       let printIt = (horseStateAndArrowOut h p k newX lastX true [] high) in
	 (print t p k x lastX)) in
    (print todo pp kk xx lastXX);;

let horsePrintDriver p k wordArray high = 
  let words = (Array.length wordArray) in
  let longestStateSequence = (findLongSequence wordArray ((Array.length wordArray)-1) 0) in
  let stateBuffer = 200 in
  let wordWidth = ((longestStateSequence * 100) + (longestStateSequence * stateBuffer) + 300) in
  let width = (wordWidth * words) in
  let newHigh = (int_of_float (high *. 1000.0)) in
  let height = (newHigh + 1000) in
  let xx = 200 in
  let lastX = 200 in
  let mh = (makeHeader p width height) in
  let ms = (makeScale p newHigh (xx-100)) in
  let rec hpd wordArray p k x lastX ww i j high = match (i < j) with
    false -> (output_string p "</svg>"); (flush p); (close_out p)
  |true ->  
     let thisWord = wordArray.(i) in
     let lastWordPP = if (i > 0) then (sumPrefixProbs wordArray.(i-1) 0.0) else (1.0) in
       let ow = (outputWord thisWord p x lastWordPP high) in
       let takenCareOf = (horsePrint thisWord p k x lastX [] high) in
       if ((List.length takenCareOf) = (List.length thisWord)) then (hpd wordArray p k (x+ww) x ww (i+1) j high) else ((printNonHistory thisWord takenCareOf p k x lastX high);(hpd wordArray p k (x+ww) x ww (i+1) j high)) in 
  try (hpd wordArray p k xx lastX wordWidth 0 (Array.length wordArray) (newHigh+500)) with Failure "hd" -> (print_string "error horsePrinterDriver");;

(*This ensures that the prefix probabilities are computed by word, not by action. *)
let findSurpPosition it = match it.config.tau with
    	[] -> (match it.arcInfo.op with 
	    Shift|Right ->(match it.config.stack with
		[] -> -2
	      |hh::tt -> ((List.hd it.config.stack).pos))
	  |_ -> -2)
       |h::t -> (match it.arcInfo.op with
	   Left|Reduce -> ((List.hd it.config.tau).pos)
	   |Shift|Right -> (match it.config.stack with
		 [] -> -2
	       |hhh::ttt ->((List.hd it.config.stack).pos)));;

let rec goThroughK kConfigs wordArray high = match kConfigs with
    [] -> high
  |h::t -> let tprob = (h.arcInfo.prob) in
    let newHigh = if (tprob > high) then (tprob) else (high) in
    let position = (findSurpPosition h) in
      if (position > (-1))
      then (let wordList = wordArray.(position) in
	    let newList = (h::wordList) in
	    (wordArray.(position) <- newList); (goThroughK t wordArray newHigh))
      else (goThroughK t wordArray newHigh);;


let rec goThroughSequence configs wordArray high = match configs with 
    [] -> high
  |h::t -> let newHigh = (goThroughK h wordArray high) in
      (goThroughSequence t wordArray newHigh);;

(*Driver*)
let statePrinterDriver configs p p2 words k = 
  let wordArray = try (Array.make (List.length (List.hd (List.hd configs)).config.tau) []) with Failure "hd" -> (Array.make 1 []) in
  let high = (goThroughSequence configs wordArray 0.) in
  let horseRace = (horsePrintDriver p2 k wordArray high) in
  let noItems = k in
  let itemWidth = noItems * 100 in
  let width = (itemWidth + (noItems * 10) + 80) in
  let height = 5000 in
  let mh = (makeHeader p width height) in
  let xx = 40 in
  let yy = 40 in
  let configs = try (List.tl configs) with Failure "tl" -> [] in
  let rec spd configs p words k x y = match configs with
    [] -> (output_string p "</svg>"); (flush p); (close_out p)
  |h::t ->  try ((statePrint h p words k x y); (spd t p words k x (y+70+70))) with Failure "hd" -> (print_string "error statePrinterDriver") in
  (spd configs p words k xx yy);;



