
 type node = {
  word : string;
  lemma : string;
  cat : string;
  pos : int;
  parent : int;
  toPeek : bool;
}

    type wordDecay = { tNode : node; timeSince : float list; fan : float; }

      module H :
  sig
    type key = Int32.t
    type 'a t = 'a Map.Make(Int32).t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
  end
 module D :
  sig
    type key = Int32.t
    type 'a t = 'a Map.Make(Int32).t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
  end

type config = {
  stack : node list;
  tau : node list;
  h : int H.t;
  d : string D.t;
  soFar : node list;
  time : float;
  decay : wordDecay list;
}

type operation = Shift | Reduce | Left | Right

type arc = {
  op : operation;
  element : string;
  prob : float;
  fromWord : node;
  toWord : node;
  fprobProb : float;
  lexicalProb : float;
  peek : bool;
}

 type model =
    Top
  | Next
  | Distance
  | Position
  | TopChildren
  | TopParent
  | TopLeft
  | TopRight
  | Stack1
  | Stack3
  | Stack2
  | Stack1Next
  | Stack2Next
  | Stack3Next
  | StackSize

type modelList = { list : model list; name : string; }

 module type FLOAT = sig type t val compare : t -> t -> int end
module Arc : sig type t = arc val compare : t -> t -> int end
module ArcSet :
  sig
    type elt = Arc.t
    type t = Set.Make(Arc).t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val for_all : (elt -> bool) -> t -> bool
    val exists : (elt -> bool) -> t -> bool
    val filter : (elt -> bool) -> t -> t
    val partition : (elt -> bool) -> t -> t * t
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
    val split : elt -> t -> t * bool * t
  end

type fp = {
  fprob : float;
  zeros : int;
  forSurprisalCalc : bool;
  lexProb : float;
}

type item = {
  config : config;
  arcInfo : arc;
  history : item list;
  rank : int;
  forwardProb : fp;
}

val makeNode : string -> string -> string -> int -> node 

val addPos : node list -> int -> node list

val makeSentence : node list -> node list 

val outputH : 'a H.t -> (H.key * 'a) list

val outputD : 'a D.t -> (D.key * 'a) list

val string_of_op : operation -> string

val natural_log : float -> float

val elementType : string ref 

val linenum : int ref 

type conditionElement = Category | Lexical | Lemma

type modelType = Generative | Conditional

val grain : conditionElement ref 

val conditionOn : node -> string 
val genCondModel : modelType ref 
